//$Author: behle $
//$Date: 2007-03-01 17:13:04 +0100 (Thu, 01 Mar 2007) $

// azove, Another Zero One Vertex Enumeration tool
// Copyright (C) 2006, Markus Behle
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA


#include <iostream>
#include <fstream>
#include <vector>
#include <set>
#include <stack>

#include "vebdd.hpp"


const char* VERSION = "1.1";
  
using namespace std;

int main(int argc, char** argv){
  cout<<"* azove v"<<VERSION<<", Another Zero One Vertex Enumeration tool"<<endl
      <<"* Copyright (C) 2006, Markus Behle"<<endl;

  bool error = false;  
  bool justCount = false;
  bool randomVarOrder = false;
  int outputNrOfPaths = noOutput;
  int andOrder = smallestFirst;
  
  if(argc == 1 || argc > 6)
    error = true;
  else{
    for(int i=1; i<argc-1; ++i){
      if(strcmp(argv[i], "-c")!=0
         && strcmp(argv[i], "-r")!=0
         && strcmp(argv[i], "-s")!=0
         && strcmp(argv[i], "-sr")!=0
         && strcmp(argv[i], "-l")!=0
         && strcmp(argv[i], "-lr")!=0
         && strcmp(argv[i], "-d")!=0
         && strcmp(argv[i], "-e")!=0)
        error = true;
      else if(strcmp(argv[i], "-c")==0)
        justCount = true;
      else if(strcmp(argv[i], "-r")==0)
        randomVarOrder = true;
      else if(strcmp(argv[i], "-s")==0)
        andOrder = smallestFirst;
      else if(strcmp(argv[i], "-sr")==0)
        andOrder = smallestFirstRandom;
      else if(strcmp(argv[i], "-l")==0)
        andOrder = largestFirst;
      else if(strcmp(argv[i], "-lr")==0)
        andOrder = largestFirstRandom;
      else if(strcmp(argv[i], "-d")==0)
        outputNrOfPaths = doubleValue;
      else if(strcmp(argv[i], "-e")==0)
        outputNrOfPaths = exactArbitraryPrec;
    }
  }
  
  if(error){
    cerr<<"Usage: "<<argv[0]<<" [-c] [-r] [-s/-l/-sr/-lr] [-d/-e] FILE"<<endl
        <<"       -c  Only count the number of 0/1 vertices [default: enumeration]"<<endl
        <<"       -r  Generate random variable order for the BDDs (may help in case of hard instances)"<<endl
        <<"           [default: variable order given by input]"<<endl
        <<"       Order for AND operation on BDDs:"<<endl
        <<"         -s  smallest BDDs first (faster but needs more memory) [default]"<<endl
        <<"         -sr smallest BDDs first and randomly if sizes equal"<<endl
        <<"         -l  largest BDDs first (slower but less memory consumption)"<<endl
        <<"         -lr largest BDDs first and randomly if sizes equal"<<endl
        <<"       Output number of 0/1 vertices with: (slows down computation time) [default: no output]"<<endl
        <<"         -d  double value"<<endl
        <<"         -e  exact arbitrary precision"<<endl
        <<endl
        <<"       FILE is a file in .ine format"<<endl
        <<"       statistics and timings will be written to stderr"<<endl
        <<"       output in .ext format will be written to stdout"<<endl;
    exit(1);
  }

  srand48(time(0));
  
  const char* filename = argv[argc - 1];
  veBDD vertices(filename, randomVarOrder, outputNrOfPaths, andOrder);
  vertices.buildConstraintBDDs();
  vertices.andBDDs();
  //vertices.outputDotFile("test.dot");
  //convert to ps via: dot -Tps test.dot -o test.ps
  if(!justCount)
    vertices.output01Points();
  
  return 0;
}


//-----------------------------------------------------------------------------------

veBDD::veBDD(const char* fileName, bool randomVarOrder, int outputNrOfPaths_, int andOrder_) : 
  outputNrOfPaths(outputNrOfPaths_),
  mgr(),
  prQueue(BDDPredicate(andOrder_)),
  resultBDD(NULL){
    ifstream inp(fileName, ios::in);

  if(!inp){
    cerr<<"ERROR: Can't open "<<fileName<<endl;
    exit(1);
  }

  string command;
  bool found = false;
  while(!found){
    if(inp.eof()){
      cerr<<"ERROR: Could not find line H-representation!"<<endl;
      exit(1);
    }
    else {
      inp >> command;
      if(command == "H-representation")
        found = true;
    }
  }
  
  found = false;
  set<unsigned int> linearitiesSet;
  while(!found){
    if(inp.eof()){
      cerr<<"ERROR: Could not find line begin!"<<endl;
      exit(1);
    }
    else {
      inp >> command;
      if(command == "begin")
        found = true;
      else if(command == "linearity"){
        unsigned int nrOfLinearities;
        inp >> nrOfLinearities;
        for(unsigned int i=0; i<nrOfLinearities; ++i){
          unsigned int readUInt;
          inp>>readUInt;
          linearitiesSet.insert(readUInt - 1);
        }
      }
    }
  }
  inp >> nrOfRows;
  inp >> nrOfCols;
  --nrOfCols;
  inp >> command;
  if(command != "integer"){
    cerr<<"ERROR: Could not find numbertype integer!"<<endl;
    exit(1);
  }

  createVarOrder(randomVarOrder);
    
  constraints.resize(nrOfRows, conBDDbuilder(&mgr));
  int readInt; //read in b -a^T means b - a^T x >=0
  int readCoeff[nrOfCols];  
  
  for(unsigned int i=0; i<nrOfRows; ++i){
    set<unsigned int>::iterator findIt = linearitiesSet.find(i);
    if(findIt != linearitiesSet.end()){ //i is in linearitiesSet
      linearitiesSet.erase(findIt);
      constraints[i].sense = 'E';
      inp >> constraints[i].constant;
      for(unsigned int j=0; j<nrOfCols; ++j)
        inp >> readCoeff[j];
    }
    else{
      //convert inequalities to -b + a^T x <= 0
      constraints[i].sense = 'L';
      inp >> readInt;
      constraints[i].constant = -readInt;
      for(unsigned int j=0; j<nrOfCols; ++j){
        inp >> readInt;
        readCoeff[j] = -readInt;
      }
    }
    //readCoeff contains now all coefficients in index order    
    // we convert it to all nonzero coefficients in level order
    for(unsigned int l=0; l<nrOfCols; ++l){
      if(readCoeff[level2index[l]] != 0){
        constraints[i].nzCoeff.push_back(readCoeff[level2index[l]]);
        constraints[i].nzCoeffInd.push_back(level2index[l]);
      }
    }
  }
  inp >> command;
  if(command != "end"){
    cerr<<"ERROR: Could not find end!"<<endl;
    exit(1);
  }
}


//-----------------------------------------------------------------------------------

void veBDD::createVarOrder(bool randomVarOrder){
  int permutation[nrOfCols];
  for(unsigned int i=0; i<nrOfCols; ++i){
    permutation[i] = i;
    mgr.addVar(i);
  }
  
  if(randomVarOrder){
    srand(time(0));
    std::random_shuffle(permutation, permutation + nrOfCols);
  }

  mgr.ShuffleHeap(permutation); //set the current permutation of the variables
  
  level2index.resize(nrOfCols);
  index2level.resize(nrOfCols);
  for(unsigned int i=0; i<nrOfCols; ++i){
    const unsigned int l = mgr.ReadPerm(i); //index2level
    index2level[i] = l;
    level2index[l] = i; //level2index could be done with i = mgr.ReadInvPerm(l)
  }

  if(randomVarOrder){  
    cerr<<"Using random variable order:";
    for(unsigned int l=0; l<nrOfCols; ++l)
      cerr<<" "<<level2index[l];
    cerr<<endl;
  }
}


//-----------------------------------------------------------------------------------

void veBDD::buildConstraintBDDs(void){
  //set node limit to the maximum
  Cudd_SetMaxLive(mgr.getManager(), std::numeric_limits<unsigned int>::max());
  //read the node limit with   b.obdd_mgr.ReadMaxLive()

  cerr<<"Building "<<constraints.size()<<" BDDs with "<<nrOfCols<<" variables: (nodes in BDD";
  if(outputNrOfPaths != noOutput)
    cerr<<", number of 0/1 vertices";
  cerr<<")"<<endl;

#ifdef USE_TIMER
  timeConBDDs.start();
#endif

  unsigned int constraintCounter = 1;
  for(vector<conBDDbuilder>::iterator it = constraints.begin(); it != constraints.end(); ++it) {
    cerr<<"Constraint "<<constraintCounter++<<": ";
    ADD* constraintBDD = it->buildBDD();
    if(*constraintBDD == mgr.addOne()){
      cerr<<"redundant "<<endl;
    }
    else if(*constraintBDD == mgr.addZero()){
      cerr<<"empty"<<endl;
      //clear the queue
      while(!prQueue.empty()){
        ADD* bdd = prQueue.top().first;
        prQueue.pop();
        delete bdd;
      }
      //just push the empty BDD in the queue
      prQueue.push(BDDnc(constraintBDD, 1));
      break;
    }
    else{
      const unsigned int nodeCount = constraintBDD->nodeCount();
      cerr<<"("<<nodeCount; // (incl. leaf0 and leaf1)
      if(outputNrOfPaths == doubleValue){
        cerr<<", "<<constraintBDD->CountMinterm(nrOfCols);      
      }
      else if(outputNrOfPaths == exactArbitraryPrec){
        cerr<<", ";
        preciseSatCount = constraintBDD->ApaCountMinterm(nrOfCols, &preciseSatCountDigits);
        mgr.ApaPrintDecimal(stderr, preciseSatCountDigits, preciseSatCount);
      }
      cerr<<")"<<endl;
      
      prQueue.push(BDDnc(constraintBDD, nodeCount));
    }
  }
  
#ifdef USE_TIMER
  timeConBDDs.stop();
#endif
}


//-----------------------------------------------------------------------------------


void veBDD::andBDDs(){
  cerr<<"AND operation on "<<prQueue.size()<<" BDDs: (nodes in BDD";
  if(outputNrOfPaths != noOutput)
    cerr<<", number of 0/1 vertices";
  cerr<<")"<<endl;

#ifdef USE_TIMER
  timeAndBDD.start();
#endif

  while(prQueue.size() > 1){
    cerr<<prQueue.size()<<" BDDs remaining: ";
    
    ADD* add1 = prQueue.top().first;
    prQueue.pop();

    unsigned int nodeCount = add1->nodeCount();
    cerr<<"("<<nodeCount;
    if(outputNrOfPaths == doubleValue){
      cerr<<", "<<add1->CountMinterm(nrOfCols);
    }
    else if(outputNrOfPaths == exactArbitraryPrec){
      cerr<<", ";
      preciseSatCount = add1->ApaCountMinterm(nrOfCols, &preciseSatCountDigits);
      mgr.ApaPrintDecimal(stderr, preciseSatCountDigits, preciseSatCount);
    }
    cerr<<") & ";
    
    ADD* add2 = prQueue.top().first;
    prQueue.pop();

    nodeCount = add2->nodeCount();
    cerr<<"("<<nodeCount;
    if(outputNrOfPaths == doubleValue){
      cerr<<", "<<add2->CountMinterm(nrOfCols);
    }
    else if(outputNrOfPaths == exactArbitraryPrec){
      cerr<<", ";
      preciseSatCount = add2->ApaCountMinterm(nrOfCols, &preciseSatCountDigits);
      mgr.ApaPrintDecimal(stderr, preciseSatCountDigits, preciseSatCount);
    }
    cerr<<") ==> ";

    ADD* addResult = new ADD((*add1) & (*add2));

    nodeCount = addResult->nodeCount();
    cerr<<"("<<nodeCount;
    if(outputNrOfPaths == doubleValue){
      cerr<<", "<<addResult->CountMinterm(nrOfCols);          
    }
    else if(outputNrOfPaths == exactArbitraryPrec){
      cerr<<", ";
      preciseSatCount = addResult->ApaCountMinterm(nrOfCols, &preciseSatCountDigits);
      mgr.ApaPrintDecimal(stderr, preciseSatCountDigits, preciseSatCount);
    }
    cerr<<") total nodes = "<<mgr.ReadNodeCount()<<endl;

    delete add1;
    delete add2;   
  
    prQueue.push(BDDnc(addResult, addResult->nodeCount()));
  }

  resultBDD = prQueue.top().first;
  prQueue.pop();
  
  rootnode = Cudd_Regular(resultBDD->getNode());  
  leaf1node = Cudd_Regular(mgr.addOne().getNode());
  leaf0node = Cudd_Regular(mgr.addZero().getNode());

#ifdef USE_TIMER
  timeAndBDD.stop();
#endif

  const unsigned int nodeCount = resultBDD->nodeCount();
  preciseSatCount = resultBDD->ApaCountMinterm(nrOfCols, &preciseSatCountDigits);

#ifdef USE_TIMER
  cerr<<"User time for building the BDDs: "<<timeConBDDs.getUserTime()<<"s"<<endl;
  cerr<<"User time for building the final BDD: "<<timeAndBDD.getUserTime()<<"s"<<endl;
#endif
  cerr<<"Final BDD consists of "<<nodeCount<<" nodes"<<endl // (incl. leaf0 and leaf1)
      <<"Number of 0/1 vertices = ";
  mgr.ApaPrintDecimal(stderr, preciseSatCountDigits, preciseSatCount);
  cerr<<endl;
  //a double valued sat count is returned by  resultBDD->CountMinterm(nrOfCols)
  
  if(*resultBDD == mgr.addZero())
    cerr<<"BDD is empty"<<endl;
  else if(*resultBDD == mgr.addOne())
    cerr<<"BDD represents the hypercube"<<endl;
}


//-----------------------------------------------------------------------------------


void veBDD::output01Points(void){
  //it would be more efficient to start from leaf1node
  //at the moment we are enumerating "all" 0/1 points
  
#ifdef USE_TIMER
  timeOutputPoints.start();
#endif

  cout<<"V-representation"<<endl
      <<"begin"<<endl;
  mgr.ApaPrintDecimal(stdout, preciseSatCountDigits, preciseSatCount);
  cout<<" "<<nrOfCols + 1<<" integer"<<endl;

  unsigned int point[nrOfCols];
  for(unsigned int i=0; i<nrOfCols; ++i)
    point[i] = 2;

  if(*resultBDD == mgr.addOne()){
    output(point);
  }
  else if(*resultBDD != mgr.addZero()){
    stack<DdNode*> nodeStack;
    DdNode* node = rootnode;
    do{
      unsigned int typeOfEdge;
      DdNode* tail;
      if(node != NULL){
        //save this node for visiting tail1 later
        nodeStack.push(node);
        //now we have to visit tail0
        tail = Cudd_Regular(Cudd_E(node));
        typeOfEdge = 0;
      }
      else{
        node = nodeStack.top();
        nodeStack.pop();
        //now we have to visit tail1
        tail = Cudd_T(node);
        typeOfEdge = 1;
      }
  
      if(tail == leaf0node){
        node = NULL;
      }
      else{
        const DdHalfWord source_index = node->index;
        const DdHalfWord source_level = index2level[source_index];
        point[source_index] = typeOfEdge;
  
        if(tail != leaf1node){
          const DdHalfWord target_level = index2level[tail->index];
          for(DdHalfWord l = source_level + 1; l < target_level; ++l)
            point[level2index[l]] = 2;
          node = tail;
        }
        else{
          for(DdHalfWord l = source_level + 1; l < nrOfCols; ++l)
            point[level2index[l]] = 2;
          output(point);
          node = NULL;
        }
      }
    } while(!(node == NULL && nodeStack.empty()));
  }
  
  cout<<"end"<<endl;

#ifdef USE_TIMER
  timeOutputPoints.stop();
  cerr<<"User time for building the BDDs: "<<timeConBDDs.getUserTime()<<"s"<<endl;
  cerr<<"User time for building the final BDD: "<<timeAndBDD.getUserTime()<<"s"<<endl
      <<"User time for output of points:  "<<timeOutputPoints.getUserTime()<<"s"<<endl;
#endif
}


//-----------------------------------------------------------------------------------


void veBDD::output(const unsigned int* point) const{
  //output all possible points for this path
  unsigned int nrOfPossibilities = 1;
  for(unsigned int i = 0; i < nrOfCols; ++i)
    if(point[i] == 2)
      nrOfPossibilities = nrOfPossibilities<<1;  // *2
  for(unsigned int k = 0; k < nrOfPossibilities; ++k){
    unsigned int selectBit = nrOfPossibilities;
    cout<<"1";
    for(unsigned int i = 0; i < nrOfCols; ++i)
      if(point[i] == 2){
        selectBit = selectBit>>1; // /2
        cout<<" "<<((k & selectBit) ? 1 : 0);
      }
      else
        cout<<" "<<point[i];
    cout<<endl;
  }
}
